#!/tvbin/tivosh
###############################################################################
#				TivoTitle V0.94
#
# Overview
#	TivoTitle displays the list of shows from the Tivo Now Showing list,
#	and can be configured to perform a search and replace operation
#	in the Tivo's MFS database in order to make certain titles more
#	meaningful.
#
#	For example, if you have a season pass that records multiple episodes
#	of the show Behind The Music, you can't tell at a glance which
#	episode is which. TivoTitle will append the episode name to each
#	of the titles, so that you can distinguish between the episodes
#	in the Now Showing list. 
#
#	TivoTitle is a Tcl script that runs under the 'tivosh' interpreter,
#	on the Tivo receiver. TivoTitle can be run from the command prompt,
#	or you can use RunTivoTitle (a shell script provided with TivoTitle)
#	to run TivoTitle once a day so that newly recorded shows are
#	automatically renamed.
#
#	You can run TivoTitle in 'safe' mode, where it will show you the
#	database changes that it intends to make, without making them.
#
#	Note that if you are looking at the Now Showing list on the Tivo
#	when TivoTitle renames a show, it may take 1-2 minutes before the
#	Tivo redraws the list. Exiting and re-entering Now Playing will
#	not speed this up.
#
#	See 'usage' in the code below for usage information and examples.
#
# WARNING
#	TivoTitle is provided as-is without warranty or guarantee.
#	The author(s) assume no liability for any personal or property
#	damage that may occur in the process of a) preparing your Tivo for the
#	installation of TivoTitle or similar programs, b) installing TivoTitle
#	into your Tivo receiver, or c) running TivoTitle.
#
# Requirements
#	In order to install and use TivoTitle, you need need serial port access
#	to the bash prompt on your Tivo receiver, and you need to take the
#	necessary steps to ensure that files that you copy to the Tivo will
#	not be deleted by the Tivo system. Refer to the online Tivo forums
#	for information on those topics.
#
# Compatibility
#	TivoTitle has been tested on the following Tivo receiver types:
#
#	Mfr	Type	Model #		Tivo Version	TivoTitle Ver. + Result
#	------- ------- --------------- --------------- -----------------------
#	Hughes	DTivo	GXCEBOTD	2.5-01-1-031	V0.9 PASSED	1/16/2002
#       Philips	DTivo	DSR6000		2.5.2-01-1-011	V0.93 PASSED	8-20-2002
#
# Installation
#	Copy TivoTitle.tcl to your development directory on the Tivo, for
#	example to /var/hack/bin. Make it executable using 
#	'chmod 755 TivoTitle.tcl'. Then if /var/hack/bin is in your PATH
#	then you can invoke TivoTitle.tcl from any directory.
#
# Copying
#	Feel free to copy, distribute, post, modify, and/or improve TivoTitle
#	with the following conditions: 1) as long as this file is called
#	TivoTitle.tcl or shows the name "TivoTitle" in this header, this header
#	must be preserved; and 2) if you make any coding change you must
#	increment the version number appropriately, in both the title above
#	and the version variable defined below, and update the Revision
#	History section of this header. (I am not going to centrally manage
#	changes but I insist that if you make a change, you make it obvious
#	by upping the version number.)
#
# Revision History
#
# Version		Author(s)
#			Description of changes
# ----- ---------	-----------------------------------------------------
# V0.9	1/16/2002	Eric2048 (Eric E. Del Sesto, Eric2048@attbi.com)	
#			Initial version
#
# V0.91	1/19/2002	Eric2048
#			Now using the built-in ForeachMfsFile, and customized
#			usage of RetryTransaction. Some doc improvements.
#			Small change to the E! sample title.
#
# V0.92 1/26/2002	Eric2048
#			Numerous improvements to the search & replace options.
#			See the new title_table documentation for info.
#
# V0.93 8/20/2002	AlphaWolf_HK / MrBlack51
#			Added compatibility for version 3 tivos, and made an
#			attempt to reduce the skipping that some people
#			experience while TivoTitle is doing its work.
#
# V0.94 12/23/2002 MrBlack51
#			Fixed error: not closing db
#
# Other Credits
#	TivoTitle is based on example Tcl code from the tivoweb project
#	(Josha Foust et al, tivoweb@lightn.org), and from ShowTitle.tcl script
#	(Stephen Rothwell, sfr@linuxcare.com.au).
#
# Known Bugs
#	All versions: a repeat episode of a renamed show will have the new
#	title.
#
###############################################################################
# Global variables and initialization

global db

set change_mode 0
set delay_mode 0
set safe_mode 0
set title_mode 0
set quiet_mode 0
set dbPoolSize [expr 100 * 1024]

###############################################################################
#				Version & Usage
set version V0.94
set usage {
Usage:
    -l
	List mode. All show titles in Now Showing will be listed. This is
	the default mode.

    -c
	Change mode. Replaces show titles which match the patterns
	specified in TivoTitle.tcl (search for 'title_table' in TivoTitle.tcl).

    -d
        Delay mode. Use this mode to add a delay to each loop. TivoTitle will
        run much slower, but it should reduce the ammount of audio skipping
        seen while TivoTitle is running.

    -s	Safe mode. Use this in combination with -c to see the changes
	that TivoTitle would make, without actually changing the Tivo database.

    -q	Quiet mode (by default, all of the above modes are verbose).

    -h	help

Examples:

    TivoTitle.tcl
	List mode: shows list of show titles.

    TivoTitle.tcl -c -s
	Change mode and Safe mode: perform standard search and replace on show
	titles, show the result, but don't actually change the Tivo database.

    TivoTitle.tcl -c
	Change mode: perform standard search and replace on show titles.
	
    TivoTitle.tcl -c -d -q
        Change Delay and Quiet mode: This mode is for people who run TivoTitle
        regularly from a timer, its meant to reduce overhead.

In order to confirm that your search patterns in title_table are correct,
I STRONGLY SUGGEST that you follow these steps:

1) Make sure you know what TivoTitle is going to do:
	TivoTitle.tcl -c -s

2) Let TivoTitle make those changes:
	TivoTitle.tcl -c

}

###############################################################################
#				title_table
#
# Modify this table to indicate the shows which should be renamed.
#
# Format:
#
# set title_table {
#    {
#	<current title>
#	<new title>
#	<append flag>
#    }
#    [additional show definitions]
# }
#
# Each of these fields is described below.
#
#
# FIELD: <current title>
#
# This is the title to search for. You can specify an exact show title,
# for example:
#
#	Rugrats
#	"Behind the Music"
#
# or you can specify a regular expression, using any combination of:
#
#	*	match any sequence of characters, including a null string
#	?	match any single character in <title>
#	[chars] match any character in the set given by chars
#	[char-char]
#		match any character in the set given by char-char,
#		inclusive
#	\x	match any single character. Use this to embed *, ?, [, ], or \
#
# some examples:
#
#	"The Tonight Show*"
#	"Behind [Tt]he Music"
#
# Or, you can specify the special string ANY_TITLE_WITHOUT_COLON, for example:
#	ANY_TITLE_WITHOUT_COLON
#
# that will match any show title that does not have a colon (":") character.
#
# If <current title> has any embedded spaces, enclose it in "". If it has any
# embedded quotes ("), enclose it in {}.
#
#
# FIELD: <new title>
#
# This is the new title to use for any show which matches the <current title>
# that you specified. You can specify a string, for example:
#	"Rugrats: "
#
# or you can specify the command SAME_TITLE_WITH_COLON, as in:
#	SAME_TITLE_WITH_COLON
#
# That will cause TivoTitle to append a colon and a space (": ") to the show's
# current title. You use that in combination with one of the APPEND_EPISODE...
# flags (see below).
#
# If <new title> has any embedded spaces, enclose it in "". If it has any
# embedded quotes ("), enclose it in {}.
#
# Note: if <current title> is ANY_TITLE_WITHOUT_COLON, then TivoTitle ignores
# the <new title> field and assumes that you are using the
# SAME_TITLE_WITH_COLON mode. Otherwise you would be asking TivoTitle to rename
# multiple shows to the same constant title.
#
#
# FIELD: <append flag>
#
# This field must contain one and only one of the following flags:
#
#	APPEND_NOTHING	Do not append additional text to the title.
#
#	APPEND_EPISODE	Append the episode title to the title, if there is
#			an episode title.
#
#	APPEND_EPISODE_OR_DESCRIPTION
#			If there is no episode title, then append the first
#			30 characters of the description to the title.
#			(I recommend that you first try this on 1 or 2 shows
#			to make sure that you like the result.)
#
# Example #1:
#
# This is the default configuration for TivoTitle. Any show that has a title
# without a colon character, and that has an episode title, will be renamed
# to "<title>: <episode title>".
#
# set title_table {
#    {
#	ANY_TITLE_WITHOUT_COLON
#	SAME_TITLE_WITH_COLON
#	APPEND_EPISODE
#    }
#
# For instance, the show titled "Behind the Music", episode title "Sting",
# would be renamed to "Behind the Music: Sting". If a particular show did not
# have an episode title, it would not be renamed.
#
# Example #2:
#
# This is a more complex do-it-yourself example.
#
# set title_table {
#    {
#	"Behind the Music"
#	SAME_TITLE_WITH_COLON
#	APPEND_EPISODE
#    }
#    {
#	"The E! True Hollywood Story"
#	"E! Story: "
#	APPEND_EPISODE_OR_DESCRIPTION
#    }
# }
#
# title_table: MODIFY THIS to customize TivoTitle

set title_table {
    {
	"CSI: Crime Scene Investigation"
	"CSI:"
	APPEND_EPISODE
    }
    {
	"That 70's Show"
	"70s Show:"
	APPEND_EPISODE
    }
    {
	"The Screen Savers"
	"TSS:"
	APPEND_EPISODE_OR_DESCRIPTION
    }
    {
	"3rd Rock From the Sun"
	"3rdRock:"
	APPEND_EPISODE
    }
    {
   "The Simpsons"
   "Simpsons:"
   APPEND_EPISODE_OR_DESCRIPTION
    }
    {
	"The Tonight Show With Jay Leno"
	"Tonight Show:"
	APPEND_EPISODE_OR_DESCRIPTION
    }
    {
   "King of the Hill"
   "King of Hill:"
   APPEND_EPISODE_OR_DESCRIPTION
    }
    {
   "The People's Court"
   "People's Court:"
   APPEND_EPISODE_OR_DESCRIPTION
    }
    {
	ANY_TITLE_WITHOUT_COLON
	SAME_TITLE_WITH_COLON
	APPEND_EPISODE_OR_DESCRIPTION
    }
}

###############################################################################
#				progress
#
# Print a string to stdout if not in Quiet Mode.

proc progress {str} {
    global quiet_mode

    if { $quiet_mode == 0 } {
        puts stdout $str
    }
}

###############################################################################
#				strim
#
# Remove leading or training {} characters and space characters from the
# strings extracted from the database.

proc strim {str} {
    return [string trim $str "\{\} "]
}

###############################################################################
#			    print_indented_line
#
# Print show information, wrapping text to an indented left margin.

proc print_indented_line { s } {
    global quiet_mode

    if { $quiet_mode == 0 } {
	set i [expr [string first ":" $s]+1]
	puts [string range $s 0 78]
	set s [string range $s 79 end]
	set w [expr 77-$i]
	while { [string length $s] > 0 } {
	    puts stdout [string range "                       " 0 $i] nonewline
	    puts "[string range $s 0 $w]"
	    set s [string range $s [expr $w+1] end]
	}   
    }
}

###############################################################################
#				scanlist
#
# Used in List mode and Change mode to scan over the list of recordings
# in the Now Showing list.

proc scanlist {} {
    global db
    global change_mode
    global delay_mode
    global safe_mode
    global quiet_mode
    global title_table
    global nowshowingpath

    set dbChanged 0

    # First lets check what tivo version we are using so that we can
    # adjust accordingly    
    RetryTransaction {
        set swsystem [db $db open /SwSystem/ACTIVE]
        set tivoswversion [dbobj $swsystem get Name]
        set versionthree [PrefixMatches "3" $tivoswversion]
        
        if { $versionthree == 1 } { set nowshowingpath {/Recording/NowShowingByTitle} }
        if { $versionthree == 0 } { set nowshowingpath {/Recording/NowShowing} }
    }
    
    # Scan through the entire Now Showing list, one entry at a time.
    ForeachMfsFile fsid name type "$nowshowingpath" "" {
    	if { $delay_mode == 1 } {
    	    # Give tivo some time to recover if we are in delay mode.
    	    progress "Delaying loop for 1 seconds."
    	    exec sleep 1
	}
	# For each file in the Now Showing list:
	if { $type == "tyDb" } {
	    # For each recording in the Now Showing list:
	    RetryTransaction {
		# Get the recording object for this Showing
		set rec [db $db openid $fsid]

		# Get the showing object for this recording
		set showing [dbobj $rec get Showing]

		# Note: there is also an "ActualShowing" object. But the
		# Showing object and the ActualShowing object seem to point to
		# the same Program object.

		# Get the program object for this showing
		set program [dbobj $showing get Program]

		# Read the Program object
		set title [strim [dbobj $program get Title]]
		set episode [strim [dbobj $program get EpisodeTitle]]
		set description [strim [dbobj $program get Description]]
	    }

	    # By default don't print or change the title, don't append.
	    set match 0
	    set new_title ""
	    set append_string ""

	    if { $change_mode == 1} {
		# Change Mode: see if this show matches any of the titles
		# that we are searching for.
		set i ""
		foreach i $title_table {
		    # Read the <current title>, <new title>, and <append flag>
		    # fields from this entry in title_table.
		    set ct [lindex $i 0]
		    set nt [lindex $i 1]
		    set af [lindex $i 2]

		    # Build the append_string according to the <append flag>.
		    set append_string ""
		    switch -exact $af {
			APPEND_EPISODE {
			    set append_string $episode
			}
			APPEND_EPISODE_OR_DESCRIPTION {
			    if { $episode == "" } {
				set append_string [string range $description 0 29]
			    } else {
				set append_string $episode
			    }
			}
			default {
			    # Default ("APPEND_NOTHING"): append nothing. 
			}
		    }

		    # Check for an ANY_TITLE_WITHOUT_COLON search
		    if { [string match "ANY_TITLE_WITHOUT_COLON" $ct] } {
			if { [string first ":" $title] == -1 } {
			    # We found a title without a colon. Assume that
			    # <new title> is SAME_TITLE_WITH_COLON, and build
			    # the new title.
			    if { $append_string != "" } {
				set new_title "$title: $append_string"
			    }
			    # Regardless of whether we are changing the title,
			    # we did find a match so exit the foreach loop
			    set match 1
			    break
			}
			exec sleep 1
		    } else {
			# See if we have an exact match
			if { [string match $ct $title] } {
			    # We found a matching title. Build the new title.
			    if { [string match "SAME_TITLE_WITH_COLON" $nt] } {
				if { $append_string != "" } {
				    set new_title "$title: $append_string"
				}
			    } else {
				set new_title "$nt$append_string"
			    }
			    # Regardless of whether we are changing the title,
			    # we did find a match so exit the foreach loop
			    set match 1
			    break
			}
		    }
		}
	    }

	    if { $match == 1 } {
		progress "Found:"
	    }

	    # Print this show title (and other info) if we are in List Mode
	    # or if we are in Change Mode and we found a match and we are not
	    # in quiet mode.

	    if { $change_mode == 0 || $match == 1 } {
		print_indented_line "Title      : $title"
		print_indented_line "Episode    : $episode"
		print_indented_line "Description: $description"
		print_indented_line "fsid       : $fsid"
	    }
     
	    if { $match == 1 } {
		# We found a match. Get the new title and append the Episode
		# Title if that was asked for.

		if { $new_title == "" } {
		    # No title change
		    progress "--> Not changing title."
		    progress ""
		    set match 0
		}
	    }

	    # If we still have a match, tell the user and make the change.

	    if { $match == 1 } {
		progress "Changing to: "
		print_indented_line "Title      : $new_title"

		if { $safe_mode == 1 } {
		    progress "--> Safe mode, not commiting change to database"
		} else {
		    progress "--> COMMITING change to database"
		    # To be safe, we will find the program object again
		    # before attempting to modify it.
		    RetryTransaction {
			set rec [db $db openid $fsid]
			set showing [dbobj $rec get Showing]
			set program [dbobj $showing get Program]
			dbobj $program set Title $new_title
		    }
		    set dbChanged 1
		}
	    }

	    # Skip a line inbetween shows
	    if { $change_mode == 0 || $match == 1 } {
		progress ""
	    }
	}
	# Loop to get the next recording
    }

    # If we changed anything, notify MyWorld in the appropriate way.
    # See /tvlib/tcl/tv/Inc.itcl for a list of these enums.

    if { $dbChanged == 1 } {
      progress "Signaling Tivo to redraw list within the next 60 seconds..."
      event send $TmkEvent::EVT_DATA_CHANGED $TmkDataChanged::RECORDINGS_ON_DISK 0
    }
}

###############################################################################
#				main
#
# Execution starts here. Parse the command line arguments first.

while { $argv != "" } {
    switch -exact -- [lindex $argv 0] {
        -h {
	    puts "TivoTitle $version"
	    puts "$usage"
	    exit 0
	}
	-l {
	    set argv [lrange $argv 1 end]
	}
	-c {
	    set change_mode 1
	    set argv [lrange $argv 1 end]
	}
	-d {
	    set delay_mode 1
    	    set argv [lrange $argv 1 end]
	   }
	-s {
	    set safe_mode 1
	    set argv [lrange $argv 1 end]
	}
	-q {
	    set quiet_mode 1
	    set argv [lrange $argv 1 end]
	}
	default {
	    puts stderr "TivoTitle: invalid argument: " nonewline
	    puts stderr "[lindex $argv 0]"
	    puts stderr "TivoTitle: use '-h' for help. Exiting."
	    exit -1
	}
    }
}

# initializing db now, due to A/V glitch issue
set db [dbopen $dbPoolSize]


# List mode or Change mode: call scanlist.
scanlist

dbclose $db
# Done.
